"""
:summary: arundeckrun package

:license: Apache License, Version 2.0
:author: Mark LaPerriere
:contact: otupman@antillion.com
:copyright: Mark LaPerriere 2015
"""
__docformat__ = "restructuredtext en"

__VERSION__ = (0, 2, 5)
VERSION = '.'.join(map(str, __VERSION__))
